/***********************************
 * IDCard utilites
 * 28/9/2002
 * The Flying Pig
 ***********************************/

//#define _RISC_OS
#define _MS_WIN

#include <stdio.h>
#include <miracl.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#if defined _RISC_OS
#include <kernel.h>
#include <sys/swis.h>
#endif

#if defined _MS_WIN
#include "windows.h"
#endif

#define  uDATA_BASE     (36)

#define  nDATA_NUM      (5)

void GenerateTemporaryKey(big bPublic, big bPrivate, FILE * hFileOut);
void Sign(char * pData, int nSave, FILE * hOutput, big bPrivate);
int  Verify(char * pData, int nSize, big bPublic, big bSigR, big bSigS);
void CreateData (big bPublic);
void CheckData (big bPublic, char * szSection);
int  FindSection(FILE * hFile, char * szSection);
void Encode(char * pData, int nSize, big bPublic, big bEncoded);
void GenerateKeyPair(void);
void AddFileSig (char * szSection, char * szFileName, FILE * hFileSig, big bPrivate);
int  VerifyFileSig(char * szSection, char * szFileName, FILE * hFileSig, big bPublic);
char KeyPress(void);
void PressKeyFinish(void);
void GSTrans(char * const szTranslated, char const * const szIn, int nLen);
void ShowSyntax(void);

static miracl * gMip;

static big bCommonP, bCommonQ, bCommonGen;

static int gPause = 1;

#if defined _RISC_OS
static char   szPortraitRun[256] = "Filer_Run %sPortrait.jpg";
#endif

#if defined _MS_WIN
HANDLE      ghStdin;
#endif

static char gszDirSep[2] = "/";
static char gszRootSep[2] = "$";

static char gszFileRoot[256] = "";

static char gszPublicKey[256];

static char *        szDataList[nDATA_NUM] =
{
  "Name",
  "Address",
  "Date of Birth",
  "Place of Birth",
  "Passport Number"
};

int main (int argc, char * * argv)
{
  big            bPublic;
  FILE *         hFileIn;
  int            bits;
  unsigned long  uSeed;
  int            nCount;
  int            nAct;

  char           szFileIn[256];
  char           szFileOut[256];
  int            nSearchForRoot;
  int            nHeight;

  printf("IDCard version 1.00 (28 Sep 2002)\n");
  printf("The Flying Pig\n");

#if defined _MS_WIN
  ghStdin = GetStdHandle(STD_INPUT_HANDLE);
#endif

  // Find the directory that we want to be in
  nSearchForRoot = 1;
  for (nCount = 0; nCount < argc; nCount++)
  {
    if (strcmp (argv[nCount], "-r") == 0)
    {
      nSearchForRoot = 0;
    }
  }

  // Set up the root
  if ((argc > 0) && nSearchForRoot)
  {
    GSTrans(szFileIn, argv[0], 256);
    nCount = strlen(szFileIn);
    hFileIn = NULL;
    nHeight = 3;
    while ((nCount > 0) && (nHeight > 0))
    {
      while ((nCount > 0) && (szFileIn[nCount] != gszDirSep[0]))
      {
        nCount--;
      }
      if (szFileIn[nCount] == gszDirSep[0])
      {
        szFileIn[nCount + 1] = 0;
        sprintf(szFileOut, "%s%s", szFileIn, "common.dss");
        hFileIn = fopen(szFileOut, "rb");
        if (hFileIn)
        {
          // We've found the file we want
          strcpy(gszFileRoot, szFileIn);
          fclose(hFileIn);
          nHeight = 0;
        }
        if (szFileIn[nCount - 1] == gszRootSep[0])
        {
          nHeight = 0;
        }
        nHeight--;
      }
      nCount--;
    }
  }

  // get common data
  hFileIn = NULL;
  while (hFileIn == NULL)
  {
    sprintf(szFileIn, "%s%s", gszFileRoot, "common.dss");
    hFileIn = fopen(szFileIn, "r");
    if (hFileIn == NULL)
    {
      printf("Enter root path:\n");
      gets(gszFileRoot);
      GSTrans(gszFileRoot, gszFileRoot, 256);
      if ((gszFileRoot[strlen(gszFileRoot)] != '/')
         && (gszFileRoot[strlen(gszFileRoot)] != '\\')
         && (gszFileRoot[strlen(gszFileRoot)] != '.'))
      {
        strcat(gszFileRoot, "/");
      }
    }
  }


  fscanf(hFileIn, "%d\n", & bits);

  gMip = mirsys(16+bits/MIRACL, 0);
  bCommonP = mirvar(0);
  bCommonQ = mirvar(0);
  bCommonGen = mirvar(0);
  bPublic = mirvar(0);

  gMip->IOBASE = 16;
  cinnum(bCommonP, hFileIn);
  cinnum(bCommonQ, hFileIn);
  cinnum(bCommonGen, hFileIn);
  gMip->IOBASE = 10;
  fclose(hFileIn);

  time((time_t *) & uSeed);
  irand(uSeed);

  sprintf(gszPublicKey, "%s%s", gszFileRoot, "public.dss");

  // Check for the -k switch
  for (nCount = 0; nCount < argc; nCount++)
  {
    if (argv[nCount][0] == '-')
    {
      switch (argv[nCount][1])
      {
        case 'k':
          if (argc < nCount + 2)
          {
            ShowSyntax();
            PressKeyFinish ();
            return 0;
          }
          strncpy(gszPublicKey, argv[nCount + 1], sizeof(gszPublicKey));
          break;
        case 'c':
          gPause = 0;
          break;
        case 'r':
          if (nCount == (argc - 1))
          {
            ShowSyntax ();
            return 0;
          }
          GSTrans(gszFileRoot, argv[nCount + 1], 256);
          if ((gszFileRoot[strlen(gszFileRoot)] != '/')
             && (gszFileRoot[strlen(gszFileRoot)] != '\\')
             && (gszFileRoot[strlen(gszFileRoot)] != '.'))
          {
            strcat(gszFileRoot, "/");
          }
          break;
        default:
          break;
      }
    }
  }

  if (strlen(gszFileRoot) > 0)
  {
    printf("Root set to %s\n", gszFileRoot);
  }
  printf("\n");

  nAct = 0;
  for (nCount = 0; nCount < argc; nCount++)
  {
    if (argv[nCount][0] == '-')
    {
      switch (argv[nCount][1])
      {
        case 'i':
          // Create data for a card
          CreateData(bPublic);
          nAct = 1;
          break;
        case 'g':
          // Generate a key pair
          GenerateKeyPair();
          nAct = 1;
          break;
        case 's':
          if (argc < nCount + 2)
          {
            ShowSyntax ();
          }
          else
          {
            CheckData(bPublic, argv[nCount + 1]);
          }
          nAct = 1;
          break;
        case 'v':
          // Verify the data on a card
          CheckData(bPublic, "");
          nAct = 1;
          break;
        case 'h':
          ShowSyntax ();
          nAct = 1;
          break;
        default:
          break;
      }
    }
  }

  if (nAct == 0)
  {
    CheckData(bPublic, "");
  }

  mirexit();
  PressKeyFinish ();
  return 0;
}

void CreateData (big bPublic)
{
  FILE *         hFileOut;
  FILE *         hFileSig;
  int            nCount;
  big            bPrivate;
  char           szInput[255];
  big            bEncoded;
  int            nDataCount;
  char           szFile[256];

  bPrivate = mirvar(0);
  bEncoded = mirvar(0);

  sprintf(szFile, "%s%s", gszFileRoot, "privdata.dss");
  hFileOut = fopen(szFile, "wb");
  if (hFileOut == NULL)
  {
    printf("File could not be opened for writing\n");
    return;
  }

  fprintf(hFileOut, "[Key]\n");

  printf("Generating temporary key...\n");
  GenerateTemporaryKey(bPublic, bPrivate, hFileOut);



  // Create the encrypted data
  for (nDataCount = 0; nDataCount < nDATA_NUM; nDataCount++)
  {
    for (nCount = 0; nCount < (int)sizeof(szInput); nCount++)
    {
      szInput[nCount] = 0;
    }
    printf("\nEnter %s:\n", szDataList[nDataCount]);
    gets(szInput);

    fprintf(hFileOut, "[%s]\n", szDataList[nDataCount]);
    Encode(szInput, (int)sizeof(szInput), bPublic, bEncoded);
    gMip->IOBASE=uDATA_BASE;
    cotnum(bEncoded, hFileOut);
  }

  fclose(hFileOut);


  // Load in the private key for signing
  printf("Please enter the filename of the main private key file:\n");
  gets(szInput);

  hFileSig = fopen(szInput, "rb");
  if (hFileSig == NULL)
  {
    printf("Could not open private key file to sign data\n");
    return;
  }
  gMip->IOBASE=uDATA_BASE;
  cinnum(bPrivate, hFileSig);
  fclose(hFileSig);
  sprintf(szFile, "%s%s", gszFileRoot, "signature.dss");
  hFileSig = fopen(szFile, "wb");
  AddFileSig ("Data", "privdata.dss", hFileSig, bPrivate);
  AddFileSig ("Portrait", "portrait.jpg", hFileSig, bPrivate);
  AddFileSig ("RISC OS Executable", "!IDCard/!RunImage", hFileSig, bPrivate);
  AddFileSig ("Windows Executable", "IDCard.exe", hFileSig, bPrivate);
  AddFileSig ("Public Key", "public.dss", hFileSig, bPrivate);
  AddFileSig ("Common Data", "common.dss", hFileSig, bPrivate);
  fclose(hFileSig);

  sprintf(szFile, "%s%s", gszFileRoot, "sigsig.dss");
  hFileSig = fopen(szFile, "wb");
  AddFileSig ("Signature", "signature.dss", hFileSig, bPrivate);
  fclose(hFileSig);
}

void CheckData (big bPublic, char * szSection)
{
  char           szInput[255];
  char           szFile[256];
  FILE *         hFileIn;
  FILE *         hFileSig;
  big            bSigR;
  big            bSigS;
  int            nCount;
  big            bEncoded;
  big            bVerify;
  int            nDataCount;
  int            nDataBreak;
  char           cKeyPress;
  int            nVerify;
#if defined _RISC_OS
  _kernel_swi_regs sRegisters;
  int            nCarry;
#endif

  bSigR = mirvar(0);
  bSigS = mirvar(0);
  bEncoded = mirvar(0);
  bVerify = mirvar(0);

  // Read in public key from file
  hFileIn = fopen(gszPublicKey, "rb");
  if (hFileIn == NULL)
  {
    printf("File could not be opened for reading\n");
    return;
  }
  gMip->IOBASE=uDATA_BASE;
  cinnum(bPublic, hFileIn);
  fclose(hFileIn);

  // Load the data signature from the signature file
  nVerify = 1;
  sprintf(szFile, "%s%s", gszFileRoot, "signature.dss");
  hFileSig = fopen(szFile, "rb");
  nVerify &= VerifyFileSig("Data", "privdata.dss", hFileSig, bPublic);
  nVerify &= VerifyFileSig("Portrait", "portrait.jpg", hFileSig, bPublic);
  nVerify &= VerifyFileSig("RISC OS Executable", "!IDCard/!RunImage", hFileSig, bPublic);
  nVerify &= VerifyFileSig("Windows Executable", "IDCard.exe", hFileSig, bPublic);
  nVerify &= VerifyFileSig("Public Key", "public.dss", hFileSig, bPublic);
  nVerify &= VerifyFileSig("Common Data", "common.dss", hFileSig, bPublic);
  fclose(hFileSig);

  sprintf(szFile, "%s%s", gszFileRoot, "sigsig.dss");
  hFileSig = fopen(szFile, "rb");
  nVerify &= VerifyFileSig("Signature", "signature.dss", hFileSig, bPublic);
  fclose(hFileSig);
  if (nVerify == 0)
  {
    printf("     WARNING: Verification was UNSUCCESSFUL\n");
  }

  sprintf(szFile, "%s%s", gszFileRoot, "privdata.dss");
  hFileIn = fopen(szFile, "rb");
  if (hFileIn == NULL)
  {
    printf("File could not be opened for reading\n");
    return;
  }
  FindSection(hFileIn, "Key");
  gMip->IOBASE=uDATA_BASE;
  cinnum(bPublic, hFileIn);

  if (strlen(szSection) != 0)
  {
    for (nCount = 0; nCount < (int)sizeof(szInput); nCount++)
    {
      szInput[nCount] = 0;
    }

    // Enter user data to verify
    printf("\nEnter %s:\n", szSection);
    gets(szInput);

    // Read in encoded version of name
    FindSection(hFileIn, szSection);
    gMip->IOBASE=uDATA_BASE;
    cinnum(bEncoded, hFileIn);

    // Verify data
    Encode(szInput, (int)sizeof(szInput), bPublic, bVerify);

    if (compare(bEncoded, bVerify) == 0)
    {
      printf("     Verification was SUCCESSFUL\n");
    }
    else
    {
      printf("     Verification was UNSUCCESSFUL\n");
    }
  }
  else
  {
    printf("\nPlease choose one of the following options:\n");
    for (nDataBreak = 0; nDataBreak < nDATA_NUM; nDataBreak++)
    {
      printf("  %c. Verify %s\n", nDataBreak + '1', szDataList[nDataBreak]);
    }
    printf("  %c. Display Portrait\n", nDataBreak + '1');
    printf("  %c. Verify signatures using external key\n", nDataBreak + '2');

    cKeyPress = KeyPress();
    nDataCount = cKeyPress - '1';

    if ((nDataCount < nDataBreak) && (nDataCount >= 0))
    {
      for (nCount = 0; nCount < (int)sizeof(szInput); nCount++)
      {
        szInput[nCount] = 0;
      }

      // Enter user data to verify
      printf("\nEnter %s:\n", szDataList[nDataCount]);
      gets(szInput);

      // Read in encoded version of name
      FindSection(hFileIn, szDataList[nDataCount]);
      gMip->IOBASE=uDATA_BASE;
      cinnum(bEncoded, hFileIn);

      // Verify data
      Encode(szInput, (int)sizeof(szInput), bPublic, bVerify);

      if (compare(bEncoded, bVerify) == 0)
      {
        printf("     Verification was SUCCESSFUL\n");
      }
      else
      {
        printf("     Verification was UNSUCCESSFUL\n");
      }
    }
    else
    {
      switch (nDataCount - nDataBreak)
      {
        case 0:
          // Display portrait
#if defined _RISC_OS
          sprintf(szFile, szPortraitRun, gszFileRoot);
          for (nCount = 0; nCount < (int)strlen(szFile); nCount++)
          {
            switch (szFile[nCount])
            {
              case '/':
                szFile[nCount] = '.';
                break;
              case '.':
                szFile[nCount] = '/';
                break;
              default:
                break;
            }
          }
          printf("\nDisplaying file: %s\n", szFile);
          sRegisters.r[0] = (int)szFile;
          _kernel_swi_c(OS_CLI, & sRegisters, & sRegisters, & nCarry);
#endif
#if defined _MS_WIN
		  ShellExecute(NULL,"","portrait.jpg","",gszFileRoot,SW_SHOW);
#endif
          break;
        case 1:
          // Verify data using external key
          printf("\nEnter the public key filename:\n");
          gets(szInput);
          GSTrans(szInput, szInput, 256);

          // Read in public key from file
          hFileIn = fopen(szInput, "rb");
          if (hFileIn == NULL)
          {
            printf("File could not be opened for reading\n");
            return;
          }
          gMip->IOBASE=uDATA_BASE;
          cinnum(bPublic, hFileIn);
          fclose(hFileIn);

          nVerify = 1;
          sprintf(szFile, "%s%s", gszFileRoot, "signature.dss");
          hFileSig = fopen(szFile, "rb");
          nVerify &= VerifyFileSig("Data", "privdata.dss", hFileSig, bPublic);
          nVerify &= VerifyFileSig("Portrait", "portrait.jpg", hFileSig, bPublic);
          nVerify &= VerifyFileSig("RISC OS Executable", "!IDCard/!RunImage", hFileSig, bPublic);
          nVerify &= VerifyFileSig("Windows Executable", "IDCard.exe", hFileSig, bPublic);
          nVerify &= VerifyFileSig("Public Key", "public.dss", hFileSig, bPublic);
          nVerify &= VerifyFileSig("Common Data", "common.dss", hFileSig, bPublic);
          fclose(hFileSig);

          sprintf(szFile, "%s%s", gszFileRoot, "sigsig.dss");
          hFileSig = fopen(szFile, "rb");
          nVerify &= VerifyFileSig("Signature", "signature.dss", hFileSig, bPublic);
          fclose(hFileSig);
          if (nVerify == 0)
          {
            printf("     WARNING: Verification was UNSUCCESSFUL\n");
          }
          break;
      }
    }
  }

  fclose(hFileIn);
}

////////////////////////////////////////////////////////////////////////
// GENERATE KEY
////////////////////////////////////////////////////////////////////////

void GenerateTemporaryKey(big bPublic, big bPrivate, FILE * hFileOut)
{

  powmod(bCommonGen, bCommonQ, bCommonP, bPublic);
  if (size(bPublic) != 1)
  {
    printf("Problem - generator bCommonGen is not of order bCommonQ\n");
    exit(0);
  }

/* generate public/private keys */
  bigrand(bCommonQ, bPrivate);
  powmod(bCommonGen, bPrivate, bCommonP, bPublic);
  gMip->IOBASE=uDATA_BASE;
  cotnum(bPublic, hFileOut);
//  hFileOut = fopen("private.dss", "w");
//  cotnum(bPrivate, hFileOut);
//  fclose(hFileOut);
  return;
}

void GenerateKeyPair(void)
{
  FILE *             hFile;
  big                bPublic;
  big                bPrivate;
  char               szFile[256];

  bPublic = mirvar(0);
  bPrivate = mirvar(0);

  powmod(bCommonGen, bCommonQ, bCommonP, bPublic);
  if (size(bPublic) != 1)
  {
    printf("Problem - generator bCommonGen is not of order bCommonQ\n");
    exit(0);
  }

/* generate public/private keys */
  bigrand(bCommonQ, bPrivate);
  powmod(bCommonGen, bPrivate, bCommonP, bPublic);
  hFile = fopen(gszPublicKey, "wb");
  printf("public key = ");
  gMip->IOBASE=uDATA_BASE;
  cotnum(bPublic, stdout);
  cotnum(bPublic, hFile);
  fclose(hFile);
  sprintf(szFile, "%s%s", gszFileRoot, "private.dss");
  hFile = fopen(szFile,"wb");
  cotnum(bPrivate, hFile);
  fclose(hFile);
  return;
}

////////////////////////////////////////////////////////////////////////
// SIGN
////////////////////////////////////////////////////////////////////////

void AddFileSig (char * szSection, char * szFileName, FILE * hFileSig, big bPrivate)
{
  char          szFile[256];
  char *        pData;
  long int      nFileSize;
  FILE *        hFileIn;

  // Generate signature
  printf("Generating signature for %s...\n", szSection);

  sprintf(szFile, "%s%s", gszFileRoot, szFileName);
  hFileIn = fopen(szFile, "rb");
  fseek(hFileIn, 0, SEEK_END);
  nFileSize = ftell(hFileIn);
  pData = calloc(1, nFileSize);
  if (pData == NULL)
  {
    printf("Could not allocate %ld bytes of memory to load file\n", nFileSize);
    return;
  }
  fseek(hFileIn, 0, SEEK_SET);
  if ((int)fread(pData, 1, nFileSize, hFileIn) != nFileSize)
  {
    printf("Could not successfully read file to create signature\n");
    return;
  }
  fclose(hFileIn);
  fprintf(hFileSig, "[%s]\n", szSection);
  Sign(pData, nFileSize, hFileSig, bPrivate);
  free(pData);
}

static void Hashing(char * pData, int nSize, big bHash)
{
  char    cHash[20];
  int     nCount;
  sha     sHash;

  shs_init(& sHash);
  for (nCount = 0; nCount < nSize; nCount++)
  {
    shs_process(& sHash, pData[nCount]);
  }
  shs_hash(& sHash, cHash);
  bytes_to_big(20, cHash, bHash);
}

void Sign(char * pData, int nSize, FILE * hOutput, big bPrivate)
{
  big     r, s, k, bHash;

  r = mirvar(0);
  s = mirvar(0);
  k = mirvar(0);
  bHash = mirvar(0);

  gMip->IOBASE = 10;

/* calculate r - this can be done offline,
   and hence amortized to almost nothing  */
  bigrand(bCommonQ, k);
  powmod(bCommonGen, k, bCommonP, r);   /* see brick.c for method to speed this up */
  divide(r, bCommonQ, bCommonQ);

/* calculate message digest */
  Hashing(pData, nSize, bHash);

/* calculate s */
  xgcd(k, bCommonQ, k, k, k);
  mad(bPrivate, r, bHash, bCommonQ, bCommonQ, s);
  mad(s, k, k, bCommonQ, bCommonQ, s);
  gMip->IOBASE=uDATA_BASE;
  cotnum(r, hOutput);
  cotnum(s, hOutput);

  return;
}

////////////////////////////////////////////////////////////////////////
// VERIFY KEY
////////////////////////////////////////////////////////////////////////

int VerifyFileSig(char * szSection, char * szFileName, FILE * hFileSig, big bPublic)
{
  char           szFile[256];
  char           szTemp[255];
  char *         pData;
  long int       nFileSize;
  big            bSigR;
  big            bSigS;
  FILE *         hFileIn;
  int            nReturn;

  sprintf(szTemp, "%s signature... ", szSection);
  printf("Verifying %-32s", szTemp);

  bSigR = mirvar(0);
  bSigS = mirvar(0);

  // Load the data signature from the signature file
  FindSection(hFileSig, szSection);
  gMip->IOBASE=uDATA_BASE;
  cinnum(bSigR, hFileSig);
  cinnum(bSigS, hFileSig);

  // Verify file key signature
  sprintf(szFile, "%s%s", gszFileRoot, szFileName);

  hFileIn = fopen(szFile, "rb");
  if (hFileIn == NULL)
  {
    printf("File could not be opened for reading\n");
    return 0;
  }
  fseek(hFileIn, 0, SEEK_END);
  nFileSize = ftell(hFileIn);
  pData = calloc(1, nFileSize);
  if (pData == NULL)
  {
    printf("Could not allocate %ld bytes of memory to load file\n", nFileSize);
    return 0;
  }
  fseek(hFileIn, 0, SEEK_SET);
  if ((int)fread(pData, 1, nFileSize, hFileIn) != nFileSize)
  {
    printf("Could not successfully read file to verify signature\n");
    return 0;
  }

  nReturn = Verify(pData, nFileSize, bPublic, bSigR, bSigS);

  free(pData);
  fclose(hFileIn);
  return nReturn;
}

int Verify(char * pData, int nSize, big bPublic, big bSigR, big bSigS)
{
  big     v, u1, u2, bHash;

  v = mirvar(0);
  u1 = mirvar(0);
  u2 = mirvar(0);
  bHash = mirvar(0);
  gMip->IOBASE = 10;

  Hashing(pData, nSize, bHash);

  if (compare(bSigR, bCommonQ) >= 0 || compare(bSigS, bCommonQ) >= 0)
  {
      printf("UNSUCCESSFUL\n");
      return 0;
  }
  xgcd(bSigS, bCommonQ, bSigS, bSigS, bSigS);
  mad(bHash, bSigS, bSigS, bCommonQ, bCommonQ, u1);
  mad(bSigR, bSigS, bSigS, bCommonQ, bCommonQ, u2);
  powmod2(bCommonGen, u1, bPublic, u2, bCommonP, v);
  divide(v, bCommonQ, bCommonQ);
  if (compare(v, bSigR) == 0)
  {
    printf("SUCCESSFUL\n");
    return 1;
  }
  else
  {
    printf("UNSUCCESSFUL\n");
    return 0;
  }
  return 0;
}

////////////////////////////////////////////////////////////////////////
// ENCODE
////////////////////////////////////////////////////////////////////////

void Encode(char * pData, int nSize, big bPublic, big bEncoded)
{  /*  encode using public key  */
  big         bHash;


  gMip = mirsys(100, 0);
  bHash = mirvar(0);

  Hashing(pData, nSize, bHash);
  power(bHash,3,bPublic,bEncoded);

  return;
}

////////////////////////////////////////////////////////////////////////
// GENERAL
////////////////////////////////////////////////////////////////////////

char KeyPress (void)
{
#if defined _RISC_OS
  _kernel_swi_regs sRegisters;
  int              nCarry;
#elif defined _MS_WIN
  DWORD            nRead;
  char             cKeyPress;
#else
  char             szInput[255];
#endif
#if defined _RISC_OS
  _kernel_swi_c(OS_ReadC, & sRegisters, & sRegisters, & nCarry);
  if (nCarry == 1)
  {
    sRegisters.r[0] = 0;
  }
  return (char)(sRegisters.r[0]);
#elif defined _MS_WIN
  SetConsoleMode(ghStdin, (ENABLE_PROCESSED_INPUT));
  ReadConsoleA(ghStdin, & cKeyPress, 1, & nRead, NULL);
  SetConsoleMode(ghStdin, (ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_PROCESSED_INPUT));
  return cKeyPress;
#else
  gets(szInput);
  return szInput[0];
#endif
}

void PressKeyFinish(void)
{
#if defined _RISC_OS
  _kernel_swi_regs sRegisters;
  int              nCarry;
#elif defined _MS_WIN
  DWORD            nRead;
  char             cKeyPress;
#else
  char             szInput[255];
#endif
  if (gPause == 1)
  {
#if defined _RISC_OS
    printf("\nPress a key to finish\n");
    _kernel_swi_c(OS_ReadC, & sRegisters, & sRegisters, & nCarry);
#elif defined _MS_WIN
    printf("\nPress a key to finish\n");
    SetConsoleMode(ghStdin, (ENABLE_PROCESSED_INPUT));
    ReadConsoleA(ghStdin, & cKeyPress, 1, & nRead, NULL);
    SetConsoleMode(ghStdin, (ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_PROCESSED_INPUT));
#else
    printf("\nPress ENTER to finish\n");
    gets(szInput);
#endif
  }
  else
  {
    printf("\nFinished\n");
  }
}

int FindSection(FILE * hFile, char * szSection)
{
  char           szString[256];
  char           szFullSection[256];
  char *         cReturn;
  int            nReturn;

  sprintf(szFullSection, "[%s]", szSection);

  fseek(hFile, 0, SEEK_SET);
  do
  {
    cReturn = fgets(szString, (int)sizeof(szString), hFile);
    szString[255] = 0;
    szString[strlen(szString) - 1] = 0;
  } while (cReturn && (strcmp(szString, szFullSection) != 0));

  if (cReturn)
  {
    nReturn = 1;
  }
  else
  {
    nReturn = 0;
  }

  return nReturn;
}

void GSTrans(char * const szTranslated, char const * const szIn, int nLen)
{
  int              nCount;
#if defined _RISC_OS
  _kernel_swi_regs sRegisters;
  int              nCarry;
  char             szTemp[256];

  sRegisters.r[0] = (int)szIn;
  sRegisters.r[1] = (int)szTemp;
  sRegisters.r[2] = nLen;
  _kernel_swi_c(OS_GSTrans, & sRegisters, & sRegisters, & nCarry);
  szTemp[sRegisters.r[2]] = 0;
  for (nCount = 0; nCount < sRegisters.r[2]; nCount++)
  {
    switch (szTemp[nCount])
    {
      case '.':
        szTemp[nCount] = '/';
        break;
      case '/':
        szTemp[nCount] = '.';
        break;
      default:
        break;
    }
  }
  strncpy(szTranslated, szTemp, nLen);
#else
  if (szTranslated != szIn)
  {
    strncpy(szTranslated, szIn, nLen);
  }
#endif
  // Remove trailing spaces
  nCount = strlen(szTranslated) - 1;
  while (nCount > 0)
  {
    if (isspace(szTranslated[nCount]))
    {
      szTranslated[nCount] = 0;
    }
    else
    {
      nCount = 0;
    }
  }
}

void ShowSyntax(void)
{
  printf("Syntax:\n");
  printf("Create [options] [-k filename]\nWhere options are optional and can be one of:\n");
  printf("  -v  Verify data\n");
  printf("  -i  Initiate data\n");
  printf("  -g  Generate Public Private key pair\n");
  printf("  -s <section>  Verify given data section\n");
  printf("  -c  Continue when complete without waiting for a keypress\n");
  printf("  -r <path>  Set the root path\n");
  printf("The optional -k switch allows the main public key to be specified\n");
  printf("\n");
}
